@@ -25,6 +25,7 @@
 #include "IOHIDDevice.h"
 #include "IOHIDElement.h"
 #include "IOHIDParserPriv.h"
+#include "IOHIDPointing.h"
 
 //===========================================================================
 // IOHIDDevice class
@@ -66,13 +67,94 @@ struct IOHIDReportHandler
 #else
 #define DLOG(fmt, args...)
 #endif
+            
+#define GetPointingNub(service) \
+            OSDynamicCast(IOHIDPointing, service)
+
+// *** GAME DEVICE HACK ***
+static SInt32 g3DGameControllerCount = 0;
+// *** END GAME DEVICE HACK ***
+
+//---------------------------------------------------------------------------
+// Static helper function that will return a new IOHIDevice depending
+// on the type of HID device.
+static IOHIDevice * CreateIOHIDeviceNub(IOService * owner, IOService * provider)
+{
+    IOHIDevice 	*nub = 0;
+    OSString 	*defaultBehavior;
+    if (owner == NULL || OSDynamicCast(IOHIPointing, provider))
+        return  0;
+    
+    defaultBehavior = OSDynamicCast(OSString, 
+                            owner->getProperty("HIDDefaultBehavior"));
+     
+    // Fixing a bug in IOUSBFamily that adds a space after the key
+    if (!defaultBehavior) {
+        defaultBehavior = OSDynamicCast(OSString, 
+                            owner->getProperty("HIDDefaultBehavior "));
+    }
+    
+    if (defaultBehavior &&
+        defaultBehavior->isEqualTo("Mouse")) {
+                            
+        nub = new IOHIDPointing;
+        
+        if (nub &&
+            (!nub->init() || 
+             !nub->attach(owner) || 
+             !nub->start(owner))) 
+        {
+            nub->release();
+            nub = 0;
+        }
+    }
+    
+    return nub;
+}
+
+//---------------------------------------------------------------------------
+// Notification handler to grab an instance of the Display Manager
+bool IOHIDDevice::_publishNotificationHandler(
+			void * target,
+			void * /* ref */,
+			IOService * newService )
+{
+    IOHIDDevice * self = (IOHIDDevice *) target;
+
+    // avoiding OSDynamicCast & dependency on graphics family
+    if( newService->metaCast("IODisplayWrangler")) {
+        if( !self->_displayManager) {
+            self->_displayManager = newService;
+            self->_displayManager->retain();
+        }
+    }
+
+    return true;
+}
+
 
 //---------------------------------------------------------------------------
 // Initialize an IOHIDDevice object.
 
 bool IOHIDDevice::init( OSDictionary * dict )
 {
-	return super::init(dict);
+    _reserved = IONew( ExpansionData, 1 );
+
+    if (!_reserved)
+        return false;
+        
+    _pointingNub = 0;
+    _displayManager = 0;
+    _publishNotify = 0;
+
+    // Create an OSSet to store client objects. Initial capacity
+    // (which can grow) is set at 2 clients.
+
+    _clientSet = OSSet::withCapacity(2);
+    if ( _clientSet == 0 )
+        return false;
+
+    return super::init(dict);
 }
 
 //---------------------------------------------------------------------------
@@ -105,8 +187,34 @@ void IOHIDDevice::free()
         IOLockFree( _elementLock );
         _elementLock = 0;
     }
+    
+    if ( _clientSet )
+    {
+        // Should not have any clients.
+        assert(_clientSet->getCount() == 0);
+        _clientSet->release();
+        _clientSet = 0;
+    }
+    
+    if (_publishNotify)
+    {
+        _publishNotify->release();
+        _publishNotify = 0;
+    }
+    
+    if (_displayManager)
+    {
+        _displayManager->release();
+        _displayManager = 0;
+    }
+    
+    if ( _reserved )
+    {        
+        IODelete( _reserved, ExpansionData, 1 );
+    }
 
-	return super::free();
+
+    return super::free();
 }
 
 //---------------------------------------------------------------------------
@@ -116,6 +224,7 @@ bool IOHIDDevice::start( IOService * provider )
 {
     IOMemoryDescriptor * reportDescriptor;
     IOReturn             ret;
+    // IOHIDPointing *	 tempNub;
 
     if ( super::start(provider) != true )
         return false;
@@ -164,11 +273,40 @@ bool IOHIDDevice::start( IOService * provider )
     if ( publishProperties(provider) != true )
         return false;
 
-    // Publish ourself to the registry and trigger client matching.
+    // Create an IOHIDevice nub
+    // This has to be done after we call publishProperties
+    // becuase we determine the nub to create based on the
+    // device's PrimaryUsage and PrimaryUsagePage
+    _pointingNub = GetPointingNub(CreateIOHIDeviceNub(this, provider));
 
+    // Add a notification to get an instance of the Display
+    // Manager.  This will allow us to tickle it upon receiveing
+    // new reports.  Only do this if the device is has a primary
+    // usage of generic desktop
+    OSNumber *primaryUsagePage = OSDynamicCast(OSNumber,
+                                    getProperty(kIOHIDPrimaryUsagePageKey));
+    OSNumber *primaryUsage = OSDynamicCast(OSNumber,
+                                    getProperty(kIOHIDPrimaryUsageKey));
+    if (primaryUsagePage && 
+       (primaryUsagePage->unsigned32BitValue() == kHIDPage_GenericDesktop)) 
+    {
+        _publishNotify = addNotification( gIOPublishNotification, 
+                            serviceMatching("IODisplayWrangler"),
+                            &IOHIDDevice::_publishNotificationHandler,
+                            this, 0 );
+    }
+    
+    // *** GAME DEVICE HACK ***
+    if ((primaryUsagePage && (primaryUsagePage->unsigned32BitValue() == 0x05)) &&
+        (primaryUsage && (primaryUsage->unsigned32BitValue() == 0x01))) {
+        OSIncrementAtomic(&g3DGameControllerCount);
+    }
+    // *** END GAME DEVICE HACK ***
+
+    // Publish ourself to the registry and trigger client matching.
     registerService();
 
-	return true;
+    return true;
 }
 
 //---------------------------------------------------------------------------
@@ -176,6 +314,18 @@ bool IOHIDDevice::start( IOService * provider )
 
 void IOHIDDevice::stop(IOService * provider)
 {
+    // *** GAME DEVICE HACK ***
+    OSNumber *primaryUsagePage = OSDynamicCast(OSNumber,
+                                    getProperty(kIOHIDPrimaryUsagePageKey));
+    OSNumber *primaryUsage = OSDynamicCast(OSNumber,
+                                    getProperty(kIOHIDPrimaryUsageKey));
+                                    
+    if ((primaryUsagePage && (primaryUsagePage->unsigned32BitValue() == 0x05)) &&
+        (primaryUsage && (primaryUsage->unsigned32BitValue() == 0x01))) {
+        OSDecrementAtomic(&g3DGameControllerCount);
+    }
+    // *** END GAME DEVICE HACK ***
+    
     handleStop(provider);
 
     if ( _elementLock )
@@ -184,10 +334,84 @@ void IOHIDDevice::stop(IOService * provider)
         _readyForInputReports = false;
         ELEMENT_UNLOCK;
     }
+    
+    if ( _pointingNub ) {
+    
+        _pointingNub->stop(this);
+        _pointingNub->detach(this);
+        
+        _pointingNub->release();
+        _pointingNub = 0;
+    }
+    
 
-	super::stop(provider);
+
+    super::stop(provider);
 }
 
+//---------------------------------------------------------------------------
+// Compare the properties in the supplied table to this object's properties.
+
+static bool CompareProperty( IOService * owner, OSDictionary * matching, const char * key )
+{
+    // We return success if we match the key in the dictionary with the key in
+    // the property table, or if the prop isn't present
+    //
+    OSObject 	* value;
+    bool	matches;
+    
+    value = matching->getObject( key );
+
+    if( value)
+        matches = value->isEqualTo( owner->getProperty( key ));
+    else
+        matches = true;
+
+    return matches;
+}
+
+bool IOHIDDevice::matchPropertyTable(OSDictionary * table, SInt32 * score)
+{
+    bool match = true;
+
+    // Ask our superclass' opinion.
+    if (super::matchPropertyTable(table, score) == false)  return false;
+
+    // Compare properties.        
+    if (!CompareProperty(this, table, kIOHIDTransportKey) 	||
+        !CompareProperty(this, table, kIOHIDVendorIDKey) 	||
+        !CompareProperty(this, table, kIOHIDProductIDKey) 	||
+        !CompareProperty(this, table, kIOHIDVersionNumberKey) 	||
+        !CompareProperty(this, table, kIOHIDManufacturerKey) 	||
+        !CompareProperty(this, table, kIOHIDSerialNumberKey) 	||
+        !CompareProperty(this, table, kIOHIDLocationIDKey) 	||
+        !CompareProperty(this, table, kIOHIDPrimaryUsageKey) 	||
+        !CompareProperty(this, table, kIOHIDPrimaryUsagePageKey))
+        match = false;
+
+    // *** HACK ***
+    // RY: For games that are accidentaly matching on the keys
+    // PrimaryUsage = 0x01
+    // PrimaryUsagePage = 0x05
+    // If there no devices present that contain these values,
+    // then return true.
+    if (!match && (g3DGameControllerCount <= 0) && table) {
+        OSNumber *primaryUsage = OSDynamicCast(OSNumber, table->getObject(kIOHIDPrimaryUsageKey));
+        OSNumber *primaryUsagePage = OSDynamicCast(OSNumber, table->getObject(kIOHIDPrimaryUsagePageKey));
+
+        if ((primaryUsage && (primaryUsage->unsigned32BitValue() == 0x01)) &&
+            (primaryUsagePage && (primaryUsagePage->unsigned32BitValue() == 0x05))) {
+            match = true;
+            IOLog("IOHIDManager: It appears that an application is attempting to locate an invalid device.  A workaround is in currently in place, but will be removed after version 10.2\n");
+        }
+    }
+    // *** END HACK ***
+        
+    return match;
+}
+
+
+
 //---------------------------------------------------------------------------
 // Fetch and publish HID properties to the registry.
 
@@ -210,7 +434,13 @@ bool IOHIDDevice::publishProperties(IOService * provider)
     SET_PROP( newVersionNumber,          kIOHIDVersionNumberKey );
     SET_PROP( newManufacturerString,     kIOHIDManufacturerKey );
     SET_PROP( newProductString,          kIOHIDProductKey );
+    SET_PROP( newLocationIDNumber,       kIOHIDLocationIDKey );
+    
+    // RY: By default we publish the SerialNumber number, but if a
+    // SerialNumber string is present, overwrite that table entry.
     SET_PROP( newSerialNumber,           kIOHIDSerialNumberKey );
+    SET_PROP( newSerialNumberString,     kIOHIDSerialNumberKey );
+    
     SET_PROP( newPrimaryUsageNumber,     kIOHIDPrimaryUsageKey );
     SET_PROP( newPrimaryUsagePageNumber, kIOHIDPrimaryUsagePageKey );
 
@@ -229,6 +459,67 @@ void IOHIDDevice::handleStop(IOService * provider)
 {
 }
 
+//---------------------------------------------------------------------------
+// Handle a client open on the interface.
+
+bool IOHIDDevice::handleOpen(IOService *  client,
+                                    IOOptionBits options,
+                                    void *       argument)
+{
+    bool  accept         = false;
+
+    do {
+        // Was this object already registered as our client?
+
+        if ( _clientSet->containsObject(client) )
+        {
+            DLOG("%s: multiple opens from client %lx\n",
+                 getName(), (UInt32) client);
+            accept = true;
+            break;
+        }
+
+        // Add the new client object to our client set.
+
+        if ( _clientSet->setObject(client) == false )
+        {
+            break;
+        }
+
+        accept = true;
+    }
+    while (false);
+
+
+    return accept;
+}
+
+//---------------------------------------------------------------------------
+// Handle a client close on the interface.
+
+void IOHIDDevice::handleClose(IOService * client, IOOptionBits options)
+{
+    // Remove the object from the client OSSet.
+
+    if ( _clientSet->containsObject(client) )
+    {
+        // Remove the client from our OSSet.
+        _clientSet->removeObject(client);
+    }
+}
+
+//---------------------------------------------------------------------------
+// Query whether a client has an open on the interface.
+
+bool IOHIDDevice::handleIsOpen(const IOService * client) const
+{
+    if (client)
+        return _clientSet->containsObject(client);
+    else
+        return (_clientSet->getCount() > 0);
+}
+
+
 //---------------------------------------------------------------------------
 // Create a new user client.
 
@@ -300,11 +591,12 @@ IOReturn IOHIDDevice::handleReport( IOMemoryDescriptor * report,
     IOByteCount    reportLength;
     IOByteCount    segmentSize;
     IOReturn       ret = kIOReturnNotReady;
+    bool           changed = false;
 
     // Only input reports are currently handled.
     
-    if ( reportType != kIOHIDReportTypeInput )
-        return kIOReturnUnsupported;
+    //if ( reportType != kIOHIDReportTypeInput )
+    //    return kIOReturnUnsupported;
 
     // Get current time.
 
@@ -345,11 +637,11 @@ IOReturn IOHIDDevice::handleReport( IOMemoryDescriptor * report,
         // Get the first element in the report handler chain.
             
         element = GetHeadElement( GetReportHandlerSlot(reportID),
-                                  kIOHIDReportTypeInput );
+                                    reportType);
 
         while ( element )
         {
-            element->processReport( reportID,
+            changed |= element->processReport( reportID,
                                     reportData,
                                     reportLength << 3,
                                     &currentTime,
@@ -382,6 +674,19 @@ IOReturn IOHIDDevice::handleReport( IOMemoryDescriptor * report,
 }
 #endif
 
+    // Tickle the displayManager if any value changed and the
+    // the device is open
+    if ( changed && _displayManager && _clientSet->getCount()) 
+    {
+        _displayManager->activityTickle(0,0);
+    }
+
+    // pass the report to the IOHIDPointing nub
+    if ( _pointingNub )
+    {
+        _pointingNub->handleReport(report, options );
+    }
+
     return ret;
 }
 
@@ -470,21 +775,21 @@ IOReturn
 IOHIDDevice::createElementHierarchy( HIDPreparsedDataRef parseData )
 {
 	OSStatus   status;
-    HIDCaps    caps;
+    HIDCapabilities    caps;
     IOReturn   ret = kIOReturnNoMemory;
     bool       success;
 
     do {    
         // Get a summary of device capabilities.
 
-        status = HIDGetCaps( parseData, &caps );
+        status = HIDGetCapabilities( parseData, &caps );
         if ( status != kHIDSuccess )
         {
             ret = kIOReturnError;
             break;
         }
 
-        // Dump HIDCaps structure contents.
+        // Dump HIDCapabilities structure contents.
 
         DLOG("Report bytes: input:%ld output:%ld feature:%ld\n",
              caps.inputReportByteLength,
@@ -503,7 +808,7 @@ IOHIDDevice::createElementHierarchy( HIDPreparsedDataRef parseData )
         _maxInputReportSize    = caps.inputReportByteLength;
         _maxOutputReportSize   = caps.outputReportByteLength;
         _maxFeatureReportSize  = caps.featureReportByteLength;
-
+        
         // Create an OSArray to store all HID elements.
 
         _elementArray = OSArray::withCapacity(
@@ -751,12 +1056,12 @@ bool IOHIDDevice::createButtonElements( HIDPreparsedDataRef parseData,
                                         IOHIDElementType    elementType,
                                         UInt32              maxCount )
 {
-    OSStatus          status;
-    HIDButtonCapsPtr  buttons = 0;
-    UInt32            count   = maxCount;
-    bool              ret     = false;
-    IOHIDElement *    element;
-    IOHIDElement *    parent;
+    OSStatus          		status;
+    HIDButtonCapabilitiesPtr 	buttons = 0;
+    UInt32			count   = maxCount;
+    bool			ret     = false;
+    IOHIDElement *		element;
+    IOHIDElement *		parent;
 
     do {
         if ( maxCount == 0 )
@@ -767,11 +1072,11 @@ bool IOHIDDevice::createButtonElements( HIDPreparsedDataRef parseData,
         
         // Allocate memory to fetch all button elements from the parseData.
 
-        buttons = (HIDButtonCapsPtr) IOMalloc( maxCount *
-                                               sizeof(HIDButtonCaps) );
+        buttons = (HIDButtonCapabilitiesPtr) IOMalloc( maxCount *
+                                               sizeof(HIDButtonCapabilities) );
         if ( buttons == 0 ) break;
 
-        status = HIDGetButtonCaps( hidReportType,  /* HIDReportType    */
+        status = HIDGetButtonCapabilities( hidReportType,  /* HIDReportType    */
                                    buttons,        /* buttonCaps       */
                                    &count,         /* buttonCapsSize   */
                                    parseData );    /* preparsedDataRef */
@@ -804,7 +1109,7 @@ bool IOHIDDevice::createButtonElements( HIDPreparsedDataRef parseData,
     while ( false );
 
     if ( buttons )
-        IOFree( buttons, maxCount * sizeof(HIDButtonCaps) );
+        IOFree( buttons, maxCount * sizeof(HIDButtonCapabilities) );
     
     return ret;
 }
@@ -819,7 +1124,7 @@ bool IOHIDDevice::createValueElements( HIDPreparsedDataRef parseData,
                                        UInt32              maxCount )
 {
     OSStatus         status;
-    HIDValueCapsPtr  values = 0;
+    HIDValueCapabilitiesPtr  values = 0;
     UInt32           count  = maxCount;
     bool             ret    = false;
     IOHIDElement *   element;
@@ -834,11 +1139,11 @@ bool IOHIDDevice::createValueElements( HIDPreparsedDataRef parseData,
 
         // Allocate memory to fetch all value elements from the parseData.
 
-        values = (HIDValueCapsPtr) IOMalloc( maxCount *
-                                             sizeof(HIDValueCaps) );
+        values = (HIDValueCapabilitiesPtr) IOMalloc( maxCount *
+                                             sizeof(HIDValueCapabilities) );
         if ( values == 0 ) break;
 
-        status = HIDGetValueCaps( hidReportType,  /* HIDReportType    */
+        status = HIDGetValueCapabilities( hidReportType,  /* HIDReportType    */
                                   values,         /* valueCaps        */
                                   &count,         /* valueCapsSize    */
                                   parseData );    /* preparsedDataRef */
@@ -872,7 +1177,7 @@ bool IOHIDDevice::createValueElements( HIDPreparsedDataRef parseData,
     while ( false );
 
     if ( values )
-        IOFree( values, maxCount * sizeof(HIDValueCaps) );
+        IOFree( values, maxCount * sizeof(HIDValueCapabilities) );
     
     return ret;
 }
@@ -1083,15 +1388,259 @@ IOReturn IOHIDDevice::checkEventDelivery( IOHIDEventQueue *  queue,
     return kIOReturnSuccess;
 }
 
+#define SetCookiesTransactionState(element, cookies, count, state, index, offset) \
+    for (index = offset; index < count; index++) { 			\
+        element = GetElement(cookies[index]); 				\
+        if (element == NULL) 						\
+            continue; 							\
+        element->setTransactionState (state);				\
+    }
+
 //---------------------------------------------------------------------------
-// 
+// Update the value of the given element, by getting a report from
+// the device.  Assume that the cookieCount > 0
+
+OSMetaClassDefineReservedUsed(IOHIDDevice,  0);
+IOReturn IOHIDDevice::updateElementValues(IOHIDElementCookie *cookies, UInt32 cookieCount) {
+    IOMemoryDescriptor *	report = NULL;
+    IOHIDElement *		element = NULL;
+    IOHIDReportType		reportType;
+    IOByteCount			maxReportLength;
+    UInt8			reportID;
+    UInt32			index;
+    IOReturn			ret = kIOReturnError;
+    
+    ELEMENT_LOCK;
+    
+    SetCookiesTransactionState(element, cookies, 
+            cookieCount, kIOHIDTransactionStatePending, index, 0);
+            
+    ELEMENT_UNLOCK;
+    
+    maxReportLength = max(_maxOutputReportSize, 
+                            max(_maxFeatureReportSize, _maxInputReportSize));
+    
+    // Allocate a mem descriptor with the maxReportLength.
+    // This way, we only have to allocate one mem discriptor
+    report = IOBufferMemoryDescriptor::withCapacity(
+                            maxReportLength, kIODirectionIn);
+        
+    if (report == NULL) {
+        ret = kIOReturnNoMemory;
+        goto UPDATE_ELEMENT_CLEANUP;
+    }
+
+    // Iterate though all the elements in the 
+    // transaction.  Generate reports if needed.
+    for (index = 0; index < cookieCount; index++) {
+        element = GetElement(cookies[index]);
+        
+        if (element == NULL)
+            continue;
+            
+        if ( element->getTransactionState() 
+                != kIOHIDTransactionStatePending )
+            continue;
+                        
+        if ( !element->getReportType(&reportType) )
+            continue;
+
+        reportID = element->getReportID();
+        
+        ret = getReport(report, reportType, reportID);
+    
+        if (ret != kIOReturnSuccess)
+            break;
+            
+        // If we have a valid report, go ahead and process it.
+        ret = handleReport(report, reportType);
+        
+        if (ret != kIOReturnSuccess)
+            break;
+    }
+    
+    // release the report
+    report->release();
+
+UPDATE_ELEMENT_CLEANUP:
+
+    ELEMENT_LOCK;
+    // If needed, set the transaction state for the 
+    // remaining elements to idle.
+    SetCookiesTransactionState(element, cookies, 
+            cookieCount, kIOHIDTransactionStateIdle, index, index);
+    ELEMENT_UNLOCK;
+        
+    return ret;
+}
+
+//---------------------------------------------------------------------------
+// Post the value of the given element, by sending a report to
+// the device.  Assume that the cookieCount > 0
+OSMetaClassDefineReservedUsed(IOHIDDevice,  1);
+IOReturn IOHIDDevice::postElementValues(IOHIDElementCookie * cookies, UInt32 cookieCount) {
+    
+    OSArray			*pendingReports = NULL;
+    IOBufferMemoryDescriptor	*report = NULL;
+    IOHIDElement 		*element = NULL;
+    IOHIDElement 		*cookieElement = NULL;
+    UInt8			*reportData = NULL;
+    IOByteCount			maxReportLength = 0;
+    IOByteCount			reportLength = 0;
+    IOHIDReportType		reportType;
+    UInt8			reportID;
+    UInt32 			index;
+    IOReturn			ret = kIOReturnError;
+    
+    // Return an error if no cookies are being set
+    if (cookieCount == 0)
+        return ret;
+        
+    ELEMENT_LOCK;
+    
+    // Set the transaction state on the specified cookies
+    SetCookiesTransactionState(cookieElement, cookies, 
+            cookieCount, kIOHIDTransactionStatePending, index, 0);
+    
+    // Most times transaction will consist of items in one report
+    pendingReports = OSArray::withCapacity(1);
+    
+    if ( pendingReports == NULL ) {
+        ret = kIOReturnNoMemory;
+        goto POST_ELEMENT_CLEANUP;
+    }
+        
+    // Get the max report size
+    maxReportLength = max(_maxOutputReportSize, _maxFeatureReportSize);
+ 
+    // Iterate though all the elements in the 
+    // transaction.  Generate reports if needed. 
+    for (index = 0; index < cookieCount; index ++) {
+
+        cookieElement = GetElement(cookies[index]);
+    
+        if ( cookieElement == NULL )
+            continue;
+          
+        // Continue on to the next element if 
+        // we've already processed this one
+        if ( cookieElement->getTransactionState() 
+                != kIOHIDTransactionStatePending )
+            continue;
+            
+        if ( !cookieElement->getReportType(&reportType) )
+            continue;
+
+        
+        // Allocate a contiguous mem descriptor with the maxReportLength.
+        // This way, we only have to allocate one mem buffer.
+        report = IOBufferMemoryDescriptor::withCapacity(maxReportLength, kIODirectionOutIn, true);
+        
+        if ( report == NULL ) {
+            ret = kIOReturnNoMemory;
+            goto POST_ELEMENT_CLEANUP;
+        }
+            
+        // Obtain the buffer
+        reportData = (UInt8 *)report->getBytesNoCopy();
+                
+        reportID = cookieElement->getReportID();
+
+        // Start at the head element and iterate through
+        element = GetHeadElement(GetReportHandlerSlot(reportID), reportType);
+                
+        while ( element ) {
+            
+            element->createReport(reportID, reportData, &reportLength, &element);
+            
+            // If the reportLength was set, then this is
+            // the head element for this report
+            if ( reportLength ) {
+                report->setLength(reportLength);
+                reportLength = 0;
+            }
+                
+        }
+        
+        // If there are multiple reports, append
+        // the reportID to the first byte
+        if ( _reportCount > 1 ) 
+            reportData[0] = reportID;
+                  
+          
+        // Add the new report to the array of pending reports
+        // It will be sent to the device after the elementLock
+        // has been released
+        pendingReports->setObject(report);
+        report->release();
+    }
+
+POST_ELEMENT_CLEANUP:
+    // If needed, set the transaction state for the 
+    // remaining elements to idle.
+    SetCookiesTransactionState(cookieElement, cookies, 
+            cookieCount, kIOHIDTransactionStateIdle, index, index);
+    
+    ELEMENT_UNLOCK;
+    
+    // Now that we have formulated all the reports for this transaction,
+    // let's go ahead and post them to the device.
+    for (index = 0; index < pendingReports->getCount(); index++) {
+        report = (IOBufferMemoryDescriptor *)(pendingReports->getObject(index));
+        
+        if (report == NULL)
+            continue;
+        
+        // Send the report to the device
+        ret = setReport( report, reportType, reportID);
+        
+        if ( ret != kIOReturnSuccess )
+            break;
+    }
+
+    pendingReports->release();
+
+    return ret;
+}
+
+OSMetaClassDefineReservedUsed(IOHIDDevice,  2);
+OSString * IOHIDDevice::newSerialNumberString() const
+{
+    return 0;
+}
+
+OSMetaClassDefineReservedUsed(IOHIDDevice,  3);
+OSNumber * IOHIDDevice::newLocationIDNumber() const
+{
+    return 0;
+}
+
+//---------------------------------------------------------------------------
+// Get an async report from the device.
+
+OSMetaClassDefineReservedUsed(IOHIDDevice,  4);
+IOReturn IOHIDDevice::getReport( IOMemoryDescriptor * report,
+                                IOHIDReportType      reportType,
+                                IOOptionBits         options,
+                                UInt32               completionTimeout,
+                                IOHIDCompletion	*    completion)
+{
+    return kIOReturnUnsupported;
+}
+
+//---------------------------------------------------------------------------
+// Send an async report to the device.
+
+OSMetaClassDefineReservedUsed(IOHIDDevice,  5);
+IOReturn IOHIDDevice::setReport( IOMemoryDescriptor * report,
+                                IOHIDReportType      reportType,
+                                IOOptionBits         options,
+                                UInt32               completionTimeout,
+                                IOHIDCompletion	*    completion)
+{
+    return kIOReturnUnsupported;
+}
 
-OSMetaClassDefineReservedUnused(IOHIDDevice,  0);
-OSMetaClassDefineReservedUnused(IOHIDDevice,  1);
-OSMetaClassDefineReservedUnused(IOHIDDevice,  2);
-OSMetaClassDefineReservedUnused(IOHIDDevice,  3);
-OSMetaClassDefineReservedUnused(IOHIDDevice,  4);
-OSMetaClassDefineReservedUnused(IOHIDDevice,  5);
 OSMetaClassDefineReservedUnused(IOHIDDevice,  6);
 OSMetaClassDefineReservedUnused(IOHIDDevice,  7);
 OSMetaClassDefineReservedUnused(IOHIDDevice,  8);
